// Copyright (c) 2001 Hursh Jain (http://www.mollypages.org) 
// The Molly framework is freely distributable under the terms of an
// MIT-style license. For details, see the molly pages web site at:
// http://www.mollypages.org/. Use, modify, have fun !

package fc.util;

import java.util.*;
import java.text.*;

/** 
Useful to store thread-local calendar instances.  This class is intended
for servlets/molly pages. Instead of static get/set methods, this class
must be instantiated and the instance methods used to get/set the calendar
object. This allows multiple instances of this class in the webapp, with
each instance being able to get/set a seperate calendar. [If the methods in
this class were static, then only 1 calendar could be get/set per thread].
<p>
<font color=red><b>Each thread must remember to individually create a seperate
calendar instance and store it via the set method</b></font>. The usage idiom
is:
<blockquote>
<pre>
<font color=blue>
//WebApp has a map of ThreadLocalCalendars and also a instance variable
//with a default ThreadLocalCalendar
ThreadLocalCalendar mycal = WebApp.getThreadLocalCalendar("foo");
</font><font color=red>
if (mycal.isNull())  {
	mycal.set(Calendar.getInstance());
	}
</font><font color=blue>
Calendar cal = mycal.get();
</font>
</pre>
</blockquote>
Note, the lines in red are always needed anywhere/anytime this class is used.
(note, it just so happens that the {@link Calendar#getInstance} returns (at
least for now) a new object every time it is called, else which is why it is
being used in the example above, else we would have to manually create a new
Calendar instance instead).

*/
public final class ThreadLocalCalendar
{
public ThreadLocalCalendar()
	{ }
/*
Each get/set into the threadlocal must be seperately by each thread (the
initialValue() method is good for auto-assigning a new value but we
may need to assign a custom calendar value per thread, so we can't use
initialValue()
*/
private final ThreadLocal tlcal = new ThreadLocal();

public Calendar get()
	{
	return (Calendar) tlcal.get();
	}

public void set(Calendar cal)
	{
	tlcal.set(cal);
	}

public boolean isNull()
	{
	return tlcal.get() == null;
	}

public static void main (String args[]) throws Exception
	{ 
	final ThreadLocalCalendar cal1 = new ThreadLocalCalendar();
	
	Thread t1 = new TestThread(false, cal1);
	Thread t2 = new TestThread(false, cal1);
	t1.start();
	t2.start();
	
	for (int n = 0; n < 100; n++) {
		new TestThread(true, cal1).start();
		}
	}

static class TestThread extends Thread
	{
	boolean timing_only = false;
	ThreadLocalCalendar cal;
	TestThread(boolean timing_only, ThreadLocalCalendar cal) {
		this.timing_only = timing_only;
		this.cal = cal;
		}
		
	public void run()
		{
		//warm up
		if (cal.isNull()) {
			cal.set(Calendar.getInstance());
			}

		Watch w = new NanoWatch();
		w.start();
		if (cal.isNull()) {
			cal.set(Calendar.getInstance());
			}
		Calendar c = cal.get();
		w.stop();
		
		NumberFormat nf = NumberFormat.getNumberInstance();
		nf.setMinimumFractionDigits(2);
		nf.setMaximumFractionDigits(2);
		nf.setMinimumIntegerDigits(2);
	
		if (timing_only)
			System.out.println("[" + nf.format(w.getTime() / 1000000.00D) + " ms]");
		else
			System.out.println("[" + nf.format(w.getTime() / 1000000.00D) + " ms]" + Thread.currentThread() + "/Calendar-ID:[" + System.identityHashCode(c) + "] "+ c);
		}
	}
}